import ngsolve as ng
from ngsolve import x, y, z, curl, Mesh, HCurl
from netgen.csg import OrthoBrick, unit_cube, CSGeometry, Pnt
from math import pi, log
from array import array
import math as m
import sympy as sm

box = OrthoBrick(Pnt(0,0,0),Pnt(1,1,1)).bc("outer")
geo = CSGeometry()
geo.Add(box)

beta = 1 

exact = CoefficientFunction((0,0,sin(pi*x)))
curl_exact = CoefficientFunction((0,-pi*cos(pi*x),0))
curlcurl_exact = CoefficientFunction((0,0, pi*pi*sin(pi*x)))

src = curlcurl_exact - beta*exact

def Maxwell_FEP():
    # define trial- and test-functions
    u = fes.TrialFunction()
    v = fes.TestFunction()
    # the bilinear-form
    a = ng.BilinearForm(fes)
    a += ng.SymbolicBFI(curl(u) * curl(v)-beta*u*v)
    a.Assemble()
    # the right hand side
    f = ng.LinearForm(fes)
    f += ng.SymbolicLFI(src*v)
    f.Assemble()
    print('a.mat.nze1=',a.mat.nze)
    if fes.ndof > edges_no:    
        for i in range(edges_no):
            for j in range(edges_no,fes.ndof+1):
                if a.mat[i,j] != 0:
                    a.mat[i,j]=0
    a.mat.Update()
    print('a.mat.nze2=',a.mat.nze)
    # the solution field
    gfu = ng.GridFunction(fes)
    gfu.vec.data = a.mat.Inverse(fes.FreeDofs()) * f.vec
    # plot the solution
    ng.Draw(gfu, mesh, "gfu")
    # L2 error and HCurl error
    gfu1 = ng.GridFunction(fes)
    gfu2 = ng.GridFunction(fes)
    gfu1.vec.data=gfu.vec.data
    gfu1.vec[edges_no:]=0
    gfu2.vec.data=gfu.vec.data
    gfu2.vec[:edges_no]=0
    ng.Draw(gfu1, mesh, "gfu1")
    ng.Draw(gfu2, mesh, "gfu2")
    curl_gfu=curl(gfu)
    curl_u_gfu = curl(gfu1)
    curl_e_gfu = curl(gfu2)
    L2_u_err = ng.sqrt(ng.Integrate((gfu1-exact)*(gfu1-exact), mesh))
    HCurl_u_err = ng.sqrt(ng.Integrate(
        (gfu1-exact)*(gfu1-exact) +
        (curl_u_gfu-curl_exact)*(curl_u_gfu-curl_exact), mesh))
    # E err
    L2_e_err = ng.sqrt(ng.Integrate((gfu2)*(gfu2), mesh))
    HCurl_e_err = ng.sqrt(ng.Integrate(
        (gfu2)*(gfu2) +
        (curl_e_gfu)*(curl_e_gfu), mesh))
    L2_u_err_list.append(L2_u_err)
    HCurl_u_err_list.append(HCurl_u_err)
    L2_e_err_list.append(L2_e_err)
    HCurl_e_err_list.append(HCurl_e_err)

P_order = 1
Ref_num = 3
Int_mesh_size = 2 #1.2
L2_u_err_list = array('d', [0])
HCurl_u_err_list = array('d', [0])
L2_e_err_list = array('d', [0])
HCurl_e_err_list = array('d', [0])
Mesh_size_list = array('d', [0])

ngmesh = geo.GenerateMesh(maxh=Int_mesh_size)

for i in range(Ref_num):

    Mesh_size = Int_mesh_size/2**(i)
    Mesh_size_list.append(Mesh_size)
    if i > 0:
            ngmesh.Refine()
    mesh = Mesh(ngmesh)

    fes = HCurl(mesh, order=P_order, type1=True, dirichlet="outer")
    fes.SetOrder(TRIG,2)
    fes.SetOrder(TET,3)
    fes.Update()
    edges_no=mesh.nedge
    faces_no=mesh.nface
    elements_no=mesh.nv
    print('#ne=', mesh.ne)
    print('#nedge=', mesh.nedge)
    print('#nface=', mesh.nface)
    print('#nv=', mesh.nv)
    print ("fes.ndof =", fes.ndof)
    Maxwell_FEP()
    #ngmesh.Refine()

#=============================================================================
print("Nedelec space order= %d" % (P_order))
print("---------------------------------------------------------------------------------------------")
print("|", "  h  ", "|", "   L2_u_err   ", "|",
      "   L2_e_err   ", "|", " HCurl_u_err  ", "|", " HCurl_e_err  ", "|", " Effectivity  ", "|")
print("---------------------------------------------------------------------------------------------")
for i in range(1, Ref_num+1):
    print("|", "%1.3f" % Mesh_size_list[i],
    "|", "%014.11f" % L2_u_err_list[i], "|", "%014.11f"
    %  L2_e_err_list[i],
    "|", "%014.11f" % HCurl_u_err_list[i], "|", "%014.11f"
    % HCurl_e_err_list[i], "|", "%014.11f"
    % (HCurl_e_err_list[i]/HCurl_u_err_list[i]), "|")
print("---------------------------------------------------------------------------------------------")




